with Shardware; use Shardware;

with Ada.Text_IO; use Ada.Text_IO;
with Ada.Command_Line; use Ada.Command_Line;
with Ada.Directories;

with Interfaces; use Interfaces;

procedure sillymachine is

  F : Bin_IO.File_Type;
  FileName : String (1 .. 255) := (others => ascii.nul);
  FileNameLength : Integer := 0;
  ValidHeader : ByteArr (0 .. 15);
  Arr : ByteArr (0 .. 15);
  I : Unsigned_32 := 0;
  Value : Byte;

  ExecSize : Unsigned_32 := 0;
  MemorySize : Unsigned_32 := 0;

  PC : Unsigned_32 := 0; -- Program Counter
  Instruction : ByteArr (0 .. 15);

  RegisterU : RegArrU (0 .. 27);
  RegisterI : RegArrI (0 .. 27);
  RegisterF : RegArrF (0 .. 27);

  ZeroFlag : Boolean := False;

begin

  
  ValidHeader (0) := 6;
  ValidHeader (1) := 148;
  ValidHeader (2) := 32;
  ValidHeader (3) := 118;
  ValidHeader (4) := 101;
  ValidHeader (5) := 114;
  ValidHeader (6) := 121;
  ValidHeader (7) := 95;
  ValidHeader (8) := 115;
  ValidHeader (9) := 105;
  ValidHeader (10) := 108;
  ValidHeader (11) := 108;
  ValidHeader (12) := 121;
  ValidHeader (13) := 0;
  ValidHeader (14) := 0;
  ValidHeader (15) := 0;

  if Argument_Count = 0 then
    New_Line(1);
    Put_Line("Usage: sillymachine <sm binary file>");
    New_Line(1);
    goto THE_END;
  end if;
  
  FileNameLength := Argument (1)'Length;
  FileName(1 .. FileNameLength) := Argument (1);

  Bin_IO.Open(F, Bin_IO.In_File, FileName);
  while not Bin_IO.End_Of_File (F) loop
    Bin_IO.Read(F, Value);
    
    if ValidHeader (I) = Value then
      null;
    else
      New_Line(1);
      Put_Line("Error: Please use a valid sm binary file");
      New_Line(1);
      goto THE_END;
    end if;
    I := I + 1;
    if I = 16 then
      goto END_VALID;
    end if;

  end loop;
  << END_VALID >>
  Bin_IO.Close(F);
  --Put_Line("Header passed");

  ExecSize := Unsigned_32(Ada.Directories.Size(FileName));
  MemorySize := ExecSize;

  declare
    MemoryArr : ByteArr (0 .. MemorySize);
  begin
    for i in 0 .. 27 loop -- zero initialise registers
      RegisterI (i) := 0;
      RegisterF (i) := 0.0;
      RegisterU (i) := 0;
    end loop;

    for i in 0 .. MemorySize loop -- zero initialise memory
      MemoryArr (i) := 0;
    end loop;
    
    I := 0;
    Bin_IO.Open(F, Bin_IO.In_File, FileName);
    while not Bin_IO.End_Of_File (F) loop
      Bin_IO.Read(F, Value);
      MemoryArr(I) := Value;
      I := I + 1;
    end loop;
    Bin_IO.Close(F);

    PC := 64; -- initialise program counter to entry point
    while True loop
      for i in 0 .. 15 loop -- load the instruction
        Instruction (Unsigned_32(i)) := MemoryArr (PC + Unsigned_32(i));
      end loop;

      case MyShiftU(Instruction (0), 8) + Unsigned_32(Instruction (1)) is -- execute the instruction
        when 0 => null;                                               -- NOP
        when 1 => Trap(RegisterU, MemoryArr, MemorySize);                                               -- TRAP
        when 2 => MovU(RegisterU(Integer(Instruction (2))),             -- Move Unsigned_32 to Register
          BytesToU(TempBytes(Instruction)));
        when 3 => MovU(RegisterU(Integer(Instruction (2))),             -- Move Register to Register (Unsigned_32)
          RegisterU(Integer(Instruction (3))));
        when 4 => MovU(RegisterU(Integer(Instruction (2))),             -- Move Memory to Register (Unsigned_32)
          BytesToU(ReadMem(MemoryArr, BytesToU(TempBytes(Instruction)), 4)));
        when 5 => MovI(RegisterI(Integer(Instruction (2))),             -- Move Int to Register
          BytesToI(TempBytes(Instruction)));
        when 6 => MovI(RegisterI(Integer(Instruction (2))),             -- Move Register to Register (Int)
          RegisterI(Integer(Instruction (3))));
        when 7 => MovI(RegisterI(Integer(Instruction (2))),             -- Move Memory to Register (Integer)
          BytesToI(ReadMem(MemoryArr, BytesToU(TempBytes(Instruction)), 4)));
        when 8 => MovF(RegisterF(Integer(Instruction (2))),            -- Move Float to Register
          BytesToF(TempBytes(Instruction)));
        when 9 => MovF(RegisterF(Integer(Instruction (2))),             -- Move Register to Register (Float)
          RegisterF(Integer(Instruction (3))));
        when 10 => MovF(RegisterF(Integer(Instruction (2))),             -- Move Memory to Register (Float)
          BytesToF(ReadMem(MemoryArr, BytesToU(TempBytes(Instruction)), 4)));
        when 11 => AddU(RegisterU(Integer(Instruction (2))),             -- Add Register to Unsigned_32
          BytesToU(TempBytes(Instruction)));
        when 12 => AddU(RegisterU(Integer(Instruction (2))),             -- Add Register to Register (Unsigned_32) 
          RegisterU(Integer(Instruction (3))));
        when 13 => AddI(RegisterI(Integer(Instruction (2))),             -- Add Register to Integer
          BytesToI(TempBytes(Instruction)));
        when 14 => AddI(RegisterI(Integer(Instruction (2))),             -- Add Register to Register (Int) 
          RegisterI(Integer(Instruction (3))));
        when 15 => AddF(RegisterF(Integer(Instruction (2))),             -- Add Register to Float
          BytesToF(TempBytes(Instruction)));
        when 16 => AddF(RegisterF(Integer(Instruction (2))),             -- Add Register to Register (Float) 
          RegisterF(Integer(Instruction (3))));
        when 17 => SubU(RegisterU(Integer(Instruction(2))),             -- Subtract Unsigned_32 from Register
          BytesToU(TempBytes(Instruction)));
        when 18 => SubU(RegisterU(Integer(Instruction(2))),             -- Subtract Register from Register (Unsigned_32)
          RegisterU(Integer(Instruction (3))));
        when 19 => SubI(RegisterI(Integer(Instruction(2))),             -- Subtract Integer from Register
          BytesToI(TempBytes(Instruction)));
        when 20 => SubI(RegisterI(Integer(Instruction(2))),             -- Subtract Register from Register (Int)
          RegisterI(Integer(Instruction (3))));
        when 21 => SubF(RegisterF(Integer(Instruction(2))),             -- Subtract Float from Register
          BytesToF(TempBytes(Instruction)));
        when 22 => SubF(RegisterF(Integer(Instruction(2))),             -- Subtract Register from Register (Float)
          RegisterF(Integer(Instruction (3))));
        when 23 => MulU(RegisterU(Integer(Instruction(2))),             -- Multiply Register by Unsigned_32
          BytesToU(TempBytes(Instruction)));
        when 24 => MulU(RegisterU(Integer(Instruction(2))),             -- Multiply Register by Register (Unsigned_32)
          RegisterU(Integer(Instruction (3))));
        when 25 => MulI(RegisterI(Integer(Instruction(2))),             -- Multiply Register by Integer
          BytesToI(TempBytes(Instruction)));
        when 26 => MulI(RegisterI(Integer(Instruction(2))),             -- Multiply Register by Register (Int)
          RegisterI(Integer(Instruction (3))));
        when 27 => MulF(RegisterF(Integer(Instruction(2))),             -- Multiply Register by Float
          BytesToF(TempBytes(Instruction)));
        when 28 => MulF(RegisterF(Integer(Instruction(2))),             -- Multiply Register by Register (Float)
          RegisterF(Integer(Instruction (3))));
        when 29 => DivU(RegisterU(Integer(Instruction(2))),             -- Divide Register by Unsigned_32
          BytesToU(TempBytes(Instruction)));
        when 30 => DivU(RegisterU(Integer(Instruction(2))),             -- Divide Register from Register (Unsigned_32)
          RegisterU(Integer(Instruction (3))));
        when 31 => DivI(RegisterI(Integer(Instruction(2))),             -- Divide Register by Integer
          BytesToI(TempBytes(Instruction)));
        when 32 => DivI(RegisterI(Integer(Instruction(2))),             -- Divide Register by Register (Int)
          RegisterI(Integer(Instruction (3))));
        when 33 => DivF(RegisterF(Integer(Instruction(2))),             -- Divide Register by Float
          BytesToF(TempBytes(Instruction)));
        when 34 => DivF(RegisterF(Integer(Instruction(2))),             -- Divide Register by Register (Float)
          RegisterF(Integer(Instruction (3))));
        when 35 => ZeroFlag := CmpU(RegisterU(Integer(Instruction(2))),
          BytesToU(TempBytes(Instruction)));
        when 36 => ZeroFlag := CmpU(RegisterU(Integer(Instruction(2))),
          RegisterU(Integer(Instruction (3))));
        when 37 => ZeroFlag := CmpI(RegisterI(Integer(Instruction(2))),
          BytesToI(TempBytes(Instruction)));
        when 38 => ZeroFlag := CmpI(RegisterI(Integer(Instruction(2))),
          RegisterI(Integer(Instruction (3))));
        when 39 => ZeroFlag := CmpF(RegisterF(Integer(Instruction(2))),
          BytesToF(TempBytes(Instruction)));
        when 40 => ZeroFlag := CmpF(RegisterF(Integer(Instruction(2))),
          RegisterF(Integer(Instruction (3))));
        when 41 =>
          if ZeroFlag then
            PC := Jump(BytesToU(TempBytes(Instruction)));
          else
            null;
          end if;
        when 42 =>
          if ZeroFlag then
            null;
          else
            PC := Jump(BytesToU(TempBytes(Instruction)));
          end if;
        when 65535 => goto THE_END; -- exit opcode
        when others => null;
      end case;

      PC := PC + 16; -- increment program counter to next instruction
    end loop;

  end;

<< THE_END >>
exception
    when Name_Error => New_Line(1); Put("Error: File '"); Put(FileName); Put_Line("' does not exist!"); New_Line(1);
end sillymachine;
